/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in 
 * compliance with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2007.  All Rights Reserved.
 *
 * Protege was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu.
 *
 */

package edu.stanford.smi.protege.util.transaction.cache.impl;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import edu.stanford.smi.protege.util.transaction.cache.Cache;
import edu.stanford.smi.protege.util.transaction.cache.CacheResult;

public class RepeatableReadCache<S, V, R> implements Cache<S, V, R> {
    private Map<S, Set<V>> readOrWrittenVarsMap = new  HashMap<S, Set<V>>();
    private Cache<S, V, R> delegate;
    
    public RepeatableReadCache(Cache<S, V, R> delegate) {
        this.delegate = delegate;
    }

    public CacheResult<R> readCache(S session, V var) {
        Set<V> readOrWrittenVars = readOrWrittenVarsMap.get(session);       
        if (readOrWrittenVars != null && !readOrWrittenVars.contains(var)) {
            return new CacheResult<R>(null, false);
        }
        else {
            return delegate.readCache(session, var);
        }
    }

    public void updateCache(S session, V var) {
        markReadOrWritten(session, var);
        delegate.updateCache(session, var);
    }

    public void updateCache(S session, V var, R value) {
        markReadOrWritten(session, var);
        delegate.updateCache(session, var, value);
    }

    public void modifyCache(S session, V var) {
        markReadOrWritten(session, var);
        delegate.modifyCache(session, var);
    }

    public void modifyCache(S session, V var, R value) {
        markReadOrWritten(session, var);
        delegate.modifyCache(session, var, value);
    }

    public void delete(S session) {
        delegate.delete(session);
    }
    
    public boolean isDeleted() {
        return delegate.isDeleted();
    }

    public void startCompleteCache() {
        delegate.startCompleteCache();
    }

    public void finishCompleteCache() {
        delegate.finishCompleteCache();
    }
    
    public void abortCompleteCache() {
        delegate.abortCompleteCache();
    }
    
    public boolean isCacheComplete() {
        return delegate.isCacheComplete();
    }

    public void beginTransaction(S session) {
        if (getTransactionNesting(session) == 0) {
            readOrWrittenVarsMap.put(session, new HashSet<V>());
        }
        delegate.beginTransaction(session);
    }

    public void commitTransaction(S session) {
        delegate.commitTransaction(session);
        decrementTransaction(session);
    }

    public void rollbackTransaction(S session) {
        delegate.rollbackTransaction(session);
        decrementTransaction(session);
    }
   

    public int getTransactionNesting(S session) {
        return delegate.getTransactionNesting(session);
    }

    public void flush() {
        readOrWrittenVarsMap.clear();
        delegate.flush();
    }
    
    private void markReadOrWritten(S session, V var) {
        Set<V> readOrWrittenVars = readOrWrittenVarsMap.get(session);
        if (readOrWrittenVars != null) {
            readOrWrittenVars.add(var);
        }
    }
    
    private void decrementTransaction(S session) {
        if (getTransactionNesting(session) == 0) {
            readOrWrittenVarsMap.remove(session);
        }
    }
}
